"""Composer AJAX admin views
"""
import json

from django.urls import reverse_lazy
from django.http.response import HttpResponse, HttpResponseBadRequest

from core_composer_app.components.bucket import api as bucket_api
from core_composer_app.components.bucket.models import Bucket
from core_composer_app.components.type.models import Type
from core_composer_app.components.type_version_manager import api as type_version_manager_api
from core_composer_app.components.type_version_manager.models import TypeVersionManager
from core_composer_app.views.admin.forms import EditBucketForm
from core_main_app.commons import exceptions
from core_main_app.components.template.api import init_template_with_dependencies
from core_main_app.components.version_manager import api as version_manager_api
from core_main_app.views.admin.ajax import _get_xsd_content_from_html, _get_dependencies_dict
from core_main_app.views.common.ajax import EditObjectModalView


def delete_bucket(request):
    """Delete a bucket.

    Args:
        request:

    Returns:

    """
    try:
        bucket_id = request.POST['bucket_id']

        bucket = bucket_api.get_by_id(bucket_id)
        bucket_api.delete(bucket)
    except Exception as e:
        return HttpResponseBadRequest(str(e))

    return HttpResponse(json.dumps({}), content_type='application/javascript')


def resolve_dependencies(request):
    """Resolve import/includes to avoid local references.

    Args:
        request:

    Returns:

    """
    try:
        # Get the parameters
        name = request.POST.get('name', None)
        version_manager_id = request.POST.get('version_manager_id', '')
        filename = request.POST['filename']
        xsd_content = request.POST['xsd_content']
        schema_locations = request.POST.getlist('schemaLocations[]')
        dependencies = request.POST.getlist('dependencies[]')
        buckets = request.POST.getlist('buckets[]')

        # create new object
        type_object = Type(filename=filename, content=_get_xsd_content_from_html(xsd_content))
        init_template_with_dependencies(type_object, _get_dependencies_dict(schema_locations, dependencies))

        # get the version manager or create a new one
        if version_manager_id != '':
            type_version_manager = version_manager_api.get(version_manager_id)
        else:
            type_version_manager = TypeVersionManager(title=name)
        type_version_manager_api.insert(type_version_manager, type_object, buckets)
    except Exception as e:
        return HttpResponseBadRequest(str(e), content_type='application/javascript')

    return HttpResponse(json.dumps({}), content_type='application/javascript')


class EditBucketView(EditObjectModalView):
    form_class = EditBucketForm
    document = Bucket
    success_url = reverse_lazy("admin:core_composer_app_buckets")
    success_message = 'Label edited with success.'

    def _save(self, form):
        # Save treatment.
        try:
            self.object.label = form.cleaned_data.get('label')
            bucket_api.upsert(self.object)
        except exceptions.NotUniqueError:
            form.add_error(None, "A bucket with the same label already exists. Please choose another label.")
        except Exception as e:
            form.add_error(None, str(e))



"""Bucket api
"""
import random

from core_composer_app.components.bucket.models import Bucket
from core_main_app.commons.exceptions import ApiError


def get_by_id(bucket_id):
    """Return a bucket given its id.

    Args:
        bucket_id:

    Returns:

    """
    return Bucket.get_by_id(bucket_id)


def upsert(bucket):
    """Save or update a bucket.

    Args:
        bucket:

    Returns:

    """
    # get list of colors already present
    existing_colors = Bucket.get_colors()
    # set a color if not set
    if bucket.color is None:
        # get a unique color
        color = _get_random_hex_color()
        while color in existing_colors:
            color = _get_random_hex_color()

        bucket.color = color

    return bucket.save_object()


def get_all():
    """Return all buckets.

    Returns:

    """
    return Bucket.get_all()


def delete(bucket):
    """Delete a bucket.

    Args:
        bucket:

    Returns:

    """
    bucket.delete()


def _get_random_hex_color():
    """Return a random hexa color string.

    Returns:

    """
    return '#' + ''.join([random.choice('0123456789ABCDEF') for x in range(6)])


def add_type_to_buckets(version_manager, list_bucket_ids):
    """Add type version manager to buckets.

    Args:
        version_manager:
        list_bucket_ids:

    Returns:

    """
    if list_bucket_ids is None:
        return

    # Iterate through list of bucket ids
    for bucket_id in list_bucket_ids:
        try:
            # get the bucket using its id
            bucket = get_by_id(bucket_id)
        except:
            raise ApiError("No bucket found with the given id.")

        # add type to bucket
        bucket.types.append(version_manager)
        # update bucket
        upsert(bucket)


def update_type_buckets(version_manager, list_bucket_ids):
    """Remove type from current buckets and puts them in new list.

    Args:
        version_manager:
        list_bucket_ids:

    Returns:

    """
    # remove types from current buckets
    remove_type_from_buckets(version_manager)
    # add types to new list of buckets
    add_type_to_buckets(version_manager, list_bucket_ids)


def remove_type_from_buckets(version_manager):
    """Remove the type from all the buckets.

    Args:
        version_manager:

    Returns:

    """
    buckets = get_all()

    # Iterate through all buckets
    for bucket in buckets:
        # if type in bucket
        if version_manager in bucket.types:
            # remove type from the bucket
            bucket.types.remove(version_manager)
            # update bucket
            upsert(bucket)



""" REST views for the type version manager API
"""
from django.utils.decorators import method_decorator
from rest_framework.permissions import IsAuthenticated

from core_composer_app.components.type_version_manager import api as type_version_manager_api
from core_composer_app.rest.type_version_manager.abstract_views import AbstractTypeList
from core_main_app.rest.template_version_manager.views import AbstractTemplateVersionManagerList
from core_main_app.utils.decorators import api_staff_member_required


class GlobalTypeVersionManagerList(AbstractTemplateVersionManagerList):
    """ List all global type version managers
    """

    def get_template_version_managers(self):
        """ Get global type version managers

        Returns:

            TypeVersionManager
        """
        return type_version_manager_api.get_global_version_managers()


class UserTypeVersionManagerList(AbstractTemplateVersionManagerList):
    """ List all user type version managers
    """

    permission_classes = (IsAuthenticated, )

    def get_template_version_managers(self):
        """ Get all user type version managers

        Returns:

            TypeVersionManager
        """
        return type_version_manager_api.get_version_managers_by_user(user_id=self.request.user.id)


class UserTypeList(AbstractTypeList):
    """ Create a user type
    """

    permission_classes = (IsAuthenticated,)

    def post(self, request):
        """ Create a user type & type version manager

        Parameters:

            {
                "title": "title",
                "filename": "filename",
                "content": "<xs:schema xmlns:xs='http://www.w3.org/2001/XMLSchema'><xs:simpleType name='root'/></xs:schema>"
            }

        Note:

            "dependencies_dict": json.dumps({"schemaLocation1": "id1" ,"schemaLocation2":"id2"})

        Args:

            request: HTTP request

        Returns:

            - code: 201
              content: Type
            - code: 400
              content: Validation error / bad request
            - code: 500
              content: Internal server error
        """
        return super(UserTypeList, self).post(request)

    def get_user(self):
        """ Retrieve user from the request
        """
        return str(self.request.user.id)


class GlobalTypeList(AbstractTypeList):
    """ Create a global type
    """

    @method_decorator(api_staff_member_required())
    def post(self, request):
        """ Create a global type & type version manager

        Parameters:

            {
                "title": "title",
                "filename": "filename",
                "content": "<xs:schema xmlns:xs='http://www.w3.org/2001/XMLSchema'><xs:simpleType name='root'/></xs:schema>"
            }

        Note:

            "dependencies_dict": json.dumps({"schemaLocation1": "id1" ,"schemaLocation2":"id2"})

        Args:

            request: HTTP request

        Returns:

            - code: 201
              content: Type
            - code: 400
              content: Validation error / bad request
            - code: 500
              content: Internal server error
        """
        return super(GlobalTypeList, self).post(request)

    def get_user(self):
        """ None for global type
        """
        return None



"""Type unit tests
"""
from unittest.case import TestCase

from bson.objectid import ObjectId
from django.core import exceptions as django_exceptions
from django.test import override_settings
from mock.mock import Mock, patch

from core_composer_app.components.type import api as type_api
from core_composer_app.components.type.models import Type
from core_main_app.commons import exceptions


class TestTypeGet(TestCase):
    @patch.object(Type, 'get_by_id')
    def test_type_get_returns_type(self, mock_get_by_id):
        # Arrange
        type_filename = "Schema"
        mock_type = _create_mock_type(type_filename)

        mock_get_by_id.return_value = mock_type

        # Act
        result = type_api.get(mock_type.id)

        # Assert
        self.assertIsInstance(result, Type)

    @patch.object(Type, 'get_by_id')
    def test_template_get_raises_exception_if_object_does_not_exist(self, mock_get_by_id):
        # Arrange
        mock_absent_id = ObjectId()
        mock_get_by_id.side_effect = exceptions.DoesNotExist('')

        # Act + Assert
        with self.assertRaises(exceptions.DoesNotExist):
            type_api.get(mock_absent_id)


class TestTypeGetAll(TestCase):
    @patch.object(Type, 'get_all')
    def test_get_all_types_returns_types(self, mock_get_all):
        # Arrange
        mock_type1 = _create_mock_type()
        mock_type2 = _create_mock_type()

        mock_get_all.return_value = [mock_type1, mock_type2]

        # Act
        result = type_api.get_all()

        # Assert
        self.assertTrue(all(isinstance(item, Type) for item in result))


class TestTypeGetAllComplexType(TestCase):
    @patch.object(Type, 'get_all_complex_type')
    def test_get_all_types_returns_types(self, mock_get_all_complex_type):
        # Arrange
        mock_type1 = _create_mock_type()
        mock_type2 = _create_mock_type()

        mock_get_all_complex_type.return_value = [mock_type1, mock_type2]

        # Act
        result = type_api.get_all_complex_type()

        # Assert
        self.assertTrue(all(isinstance(item, Type) for item in result))


class TestTypeUpsert(TestCase):
    @override_settings(ROOT_URLCONF="core_main_app.urls")
    @patch.object(Type, 'save')
    def test_type_upsert_valid_returns_type(self, mock_save):
        type_object = _create_type()

        mock_save.return_value = type_object
        result = type_api.upsert(type_object)
        self.assertIsInstance(result, Type)

    @override_settings(ROOT_URLCONF="core_main_app.urls")
    @patch.object(Type, 'save')
    def test_type_upsert_invalid_filename_raises_validation_error(self, mock_save):
        type_object = _create_type(filename=1)
        mock_save.side_effect = django_exceptions.ValidationError("")
        with self.assertRaises(django_exceptions.ValidationError):
            type_api.upsert(type_object)

    @patch.object(Type, 'save')
    def test_type_upsert_invalid_core_type_raises_core_error(self, mock_save):
        type_object = _create_type(content="<schema></schema>")
        mock_save.return_value = None
        with self.assertRaises(exceptions.CoreError):
            type_api.upsert(type_object)


def _create_mock_type(filename="", content=""):
    """Returns a mock type

    Args:
        filename:
        content:

    Returns:

    """
    mock_type = Mock(spec=Type)
    mock_type.filename = filename
    mock_type.content = content
    mock_type.id = ObjectId()
    return mock_type


def _create_type(filename="", content=None):
    """Returns a type

    Args:
        filename:
        content:

    Returns:

    """
    if content is None:
        # set a valid content
        content = "<schema xmlns='http://www.w3.org/2001/XMLSchema'><simpleType name='type'>" \
                  "<restriction base='string'><enumeration value='test'/></restriction>" \
                  "</simpleType></schema>"
    return Type(
        id=ObjectId(),
        filename=filename,
        content=content
    )



"""Serializers used throughout the Rest API
"""
from core_composer_app.components.type.models import Type
from core_composer_app.components.type_version_manager import api as type_version_manager_api
from core_composer_app.components.type_version_manager.models import TypeVersionManager
from core_main_app.components.template.api import init_template_with_dependencies
from core_main_app.rest.template.serializers import TemplateSerializer
from core_main_app.rest.template_version_manager.serializers import TemplateVersionManagerSerializer
from core_main_app.rest.template_version_manager.utils import load_dependencies


class TypeVersionManagerSerializer(TemplateVersionManagerSerializer):
    """
        Type Version Manager serializer
    """

    def create(self, validated_data):
        """ Create.

        Args:
            validated_data:

        Returns:

        """
        return TypeVersionManager(**validated_data)


class CreateTypeSerializer(TemplateSerializer):
    """
        Create Type Version Manager serializer
    """
    def create(self, validated_data):
        """
        Create and return a new `Type` instance, given the validated data.
        """
        type_object = Type(filename=validated_data['filename'],
                           content=validated_data['content'])
        type_version_manager_object = validated_data['type_version_manager']

        # load dependencies
        dependencies_dict = load_dependencies(validated_data)

        # Update the content of the template with dependencies
        init_template_with_dependencies(type_object, dependencies_dict)

        # Create the template and its template version manager
        type_version_manager_api.insert(type_version_manager_object, type_object)

        return type_object

    def update(self, instance, validated_data):
        raise NotImplementedError("Type Version Manager should only be updated using specialized APIs.")



"""Composer app user views
"""
from os.path import join

from django.conf import settings
from django.contrib.staticfiles import finders
from django.urls import reverse_lazy

from core_composer_app.components.bucket import api as bucket_api
from core_composer_app.components.type_version_manager import api as type_version_manager_api
from core_composer_app.permissions import rights
from core_main_app.components.template import api as template_api
from core_main_app.components.template_version_manager import api as template_version_manager_api
from core_main_app.components.version_manager import api as version_manager_api
from core_main_app.utils import decorators as decorators
from core_main_app.utils.file import read_file_content, get_file_http_response
from core_main_app.utils.rendering import render
from core_main_app.utils.xml import xsl_transform
from core_main_app.views.user.views import get_context_manage_template_versions
from xml_utils.commons.constants import LXML_SCHEMA_NAMESPACE
from xml_utils.xsd_tree.operations.annotation import remove_annotations
from xml_utils.xsd_tree.xsd_tree import XSDTree
from xml_utils.xsd_types.xsd_types import get_xsd_types


# TODO: see if sessions are problematic


@decorators.permission_required(content_type=rights.composer_content_type,
                                permission=rights.composer_access, login_url=reverse_lazy("core_main_app_login"))
def index(request):
    """ Page that allows to select a template to start composing.

    Args:
        request:

    Returns:

    """
    assets = {
        "js": [],
        "css": []
    }

    global_active_template_list = template_version_manager_api.get_active_global_version_manager(_cls=True)
    user_active_template_list = template_version_manager_api.get_active_version_manager_by_user_id(request.user.id,
                                                                                                   _cls=True)

    global_active_type_list = type_version_manager_api.get_active_global_version_manager()
    user_active_type_list = type_version_manager_api.get_active_version_manager_by_user_id(request.user.id)

    context = {
        'global_templates': global_active_template_list,
        'global_types': global_active_type_list,
        'user_templates': user_active_template_list,
        'user_types': user_active_type_list,
    }

    return render(request,
                  'core_composer_app/user/index.html',
                  assets=assets,
                  context=context)


@decorators.permission_required(content_type=rights.composer_content_type,
                                permission=rights.composer_access, login_url=reverse_lazy("core_main_app_login"))
def build_template(request, template_id):
    """View that allows to build the Template.

    Args:
        request:
        template_id:

    Returns:

    """
    if template_id == "new":
        base_template_path = finders.find(join('core_composer_app', 'user', 'xsd', 'new_base_template.xsd'))
        xsd_string = read_file_content(base_template_path)
    else:
        template = template_api.get(template_id)
        xsd_string = template.content

    request.session['newXmlTemplateCompose'] = xsd_string
    request.session['includedTypesCompose'] = []

    # store the current includes/imports
    xsd_tree = XSDTree.build_tree(xsd_string)
    includes = xsd_tree.findall("{}include".format(LXML_SCHEMA_NAMESPACE))
    for el_include in includes:
        if 'schemaLocation' in el_include.attrib:
            request.session['includedTypesCompose'].append(el_include.attrib['schemaLocation'])
    imports = xsd_tree.findall("{}import".format(LXML_SCHEMA_NAMESPACE))
    for el_import in imports:
        if 'schemaLocation' in el_import.attrib:
            request.session['includedTypesCompose'].append(el_import.attrib['schemaLocation'])

    # remove annotations from the tree
    remove_annotations(xsd_tree)
    xsd_string = XSDTree.tostring(xsd_tree)

    # loads XSLT
    xslt_path = finders.find(join('core_composer_app', 'user', 'xsl', 'xsd2html.xsl'))
    # reads XSLT
    xslt_string = read_file_content(xslt_path)
    # transform XML to HTML
    xsd_to_html_string = xsl_transform(xsd_string, xslt_string)

    # 1) Get user defined types
    user_types = type_version_manager_api.get_version_managers_by_user(str(request.user.id))
    # 2) Get buckets
    buckets = bucket_api.get_all()

    # 3) no_buckets_types: list of types that are not assigned to a specific bucket
    no_buckets_types = type_version_manager_api.get_no_buckets_types()

    # 4) Build list of built-in types
    built_in_types = []
    for built_in_type in get_xsd_types():
        built_in_types.append({'current': 'built_in_type', 'title': built_in_type})

    assets = {
        "js": [
            {
                "path": 'core_composer_app/user/js/build_template.js',
                "is_raw": False
            },
            {
                "path": 'core_composer_app/user/js/build_template.raw.js',
                "is_raw": True
            },
            {
                "path": 'core_composer_app/user/js/xpath.js',
                "is_raw": False
            },
            {
                "path": 'core_composer_app/user/js/menus.js',
                "is_raw": False
            },
            {
                "path": 'core_composer_app/user/js/xsd_tree.js',
                "is_raw": False
            },
        ],
        "css": ['core_main_app/common/css/XMLTree.css',
                'core_composer_app/common/css/bucket.css',
                'core_composer_app/user/css/menu.css',
                'core_composer_app/user/css/style.css',
                'core_composer_app/user/css/xsd_tree.css']
    }
    context = {
        'buckets': buckets,
        'built_in_types': built_in_types,
        'no_buckets_types': no_buckets_types,
        'user_types': user_types,
        'xsd_form': xsd_to_html_string,
        'template_id': template_id,
    }

    modals = [
        'core_composer_app/user/builder/menus/sequence.html',
        'core_composer_app/user/builder/menus/element.html',
        'core_composer_app/user/builder/menus/element_root.html',

        'core_composer_app/user/builder/modals/root_type_name.html',
        'core_composer_app/user/builder/modals/element_name.html',
        'core_composer_app/user/builder/modals/insert_element.html',
        'core_composer_app/user/builder/modals/delete_element.html',
        'core_composer_app/user/builder/modals/change_type.html',
        'core_composer_app/user/builder/modals/save_template.html',
        'core_composer_app/user/builder/modals/save_type.html',
        'core_composer_app/user/builder/modals/save_success.html',
        'core_composer_app/user/builder/modals/occurrences.html',
        'core_composer_app/user/builder/modals/errors.html',
    ]

    return render(request,
                  'core_composer_app/user/build_template.html',
                  assets=assets,
                  context=context,
                  modals=modals)


@decorators.permission_required(content_type=rights.composer_content_type,
                                permission=rights.composer_access, login_url=reverse_lazy("core_main_app_login"))
def download_xsd(request):
    """Make the current XSD available for download.

    Args:
        request:

    Returns:

    """
    xsd_string = request.session['newXmlTemplateCompose']

    # return the file
    return get_file_http_response(file_content=xsd_string,
                                  file_name="schema.xsd",
                                  content_type="application/xsd",
                                  extension=".xsd")


def manage_type_versions(request, version_manager_id):
    """View that allows type versions management.

    Args:
        request:
        version_manager_id:

    Returns:

    """
    try:
        # get the version manager
        version_manager = version_manager_api.get(version_manager_id)
        context = get_context_manage_template_versions(version_manager, "Type")
        if 'core_parser_app' in settings.INSTALLED_APPS:
            context.update({"module_url": "core_composer_app_type_modules"})

        assets = {
                    "js": [
                        {
                            "path": 'core_main_app/common/js/templates/versions/set_current.js',
                            "is_raw": False
                        },
                        {
                            "path": 'core_main_app/common/js/templates/versions/restore.js',
                            "is_raw": False
                        },
                        {
                            "path": 'core_main_app/common/js/templates/versions/modals/disable.js',
                            "is_raw": False
                        }
                    ]
                }

        modals = ["core_main_app/admin/templates/versions/modals/disable.html"]

        return render(request,
                      'core_composer_app/user/types/versions.html',
                      assets=assets,
                      modals=modals,
                      context=context)
    except Exception as e:
        return render(request,
                      'core_main_app/common/commons/error.html',
                      context={'error': str(e)})



""" Initialize permissions for core composer app.
"""
import logging

from django.contrib.auth.models import Permission, Group

from core_composer_app.permissions import rights as composer_rights
from core_main_app.permissions import rights as main_rights

logger = logging.getLogger(__name__)


def init_permissions():
    """ Initialization of groups and permissions.
    """
    try:
        # Get or Create the default group
        default_group, created = Group.objects.get_or_create(name=main_rights.default_group)

        # Get composer permissions
        composer_access_perm = Permission.objects.get(codename=composer_rights.composer_access)
        composer_save_template_perm = Permission.objects.get(
            codename=composer_rights.composer_save_template
        )
        composer_save_type_perm = Permission.objects.get(
            codename=composer_rights.composer_save_type
        )

        # Add permissions to default group
        default_group.permissions.add(composer_access_perm,
                                      composer_save_template_perm,
                                      composer_save_type_perm)
    except Exception as e:
        logger.error("Impossible to init composer permissions: %s" % str(e))



""" Fixtures files for type version manager
"""
from core_composer_app.components.type.models import Type
from core_composer_app.components.type_version_manager.models import TypeVersionManager
from core_main_app.utils.integration_tests.fixture_interface import FixtureInterface


class TypeVersionManagerFixtures(FixtureInterface):
    """ Type Version Manager fixtures
    """
    type_1_1 = None
    type_1_2 = None
    type_1_3 = None
    type_2_1 = None
    type_vm_1 = None
    type_vm_2 = None
    type_vm_collection = None

    def insert_data(self):
        """ Insert a set of Types and Type Version Managers.

        Returns:

        """
        # Make a connexion with a mock database
        self.type_1_1 = Type(filename="type1_1.xsd",
                             content="content1_1",
                             hash="hash1_1",
                             is_complex=True).save()
        self.type_1_2 = Type(filename="type1_2.xsd",
                             content="content1_2",
                             hash="hash1_2",
                             is_complex=True).save()
        self.type_1_3 = Type(filename="type1_3.xsd",
                             content="content1_3",
                             hash="hash1_3",
                             is_complex=True).save()
        self.type_2_1 = Type(filename="type2_1.xsd",
                             content="content2_1",
                             hash="hash2_1",
                             is_complex=True).save()

        self.type_vm_1 = TypeVersionManager(title="type 1",
                                            user=None,
                                            versions=[str(self.type_1_1.id),
                                                      str(self.type_1_2.id),
                                                      str(self.type_1_3.id)],
                                            current=str(self.type_1_3.id),
                                            is_disabled=False,
                                            disabled_versions=[str(self.type_1_2.id)]).save()

        self.type_vm_2 = TypeVersionManager(title="type 2",
                                            user="1",
                                            versions=[str(self.type_2_1.id)],
                                            current=str(self.type_2_1.id),
                                            is_disabled=False,
                                            disabled_versions=[]).save()

        self.type_vm_collection = [self.type_vm_1, self.type_vm_2]



""" Url router for the composer application
"""
from django.conf import settings
from django.conf.urls import include
from django.urls import re_path

from core_composer_app.views.user import views as user_views, ajax as user_ajax

parser_url = []
if 'core_parser_app' in settings.INSTALLED_APPS:
    from core_parser_app.views.common import views as common_parser_views

    parser_url = [
        re_path(r'^type/modules/(?P<pk>\w+)',
                common_parser_views.ManageModulesUserView.as_view(
                    back_to_previous_url="core_composer_app_manage_type_versions"
                ),
                name='core_composer_app_type_modules'),
    ]

urlpatterns = [
    re_path(r'^$', user_views.index,
            name='core_composer_index'),
    re_path(r'^build-template/(?P<template_id>\w+)$', user_views.build_template,
            name='core_composer_build_template'),
    re_path(r'^download-xsd$', user_views.download_xsd,
            name='core_composer_download_xsd'),
    re_path(r'^type/versions/(?P<version_manager_id>\w+)', user_views.manage_type_versions,
            name='core_composer_app_manage_type_versions'),
    re_path(r'^change-xsd-type$', user_ajax.change_xsd_type,
            name='core_composer_change_xsd_type'),
    re_path(r'^change-root-type-name$', user_ajax.change_root_type_name,
            name='core_composer_change_root_type_name'),
    re_path(r'^insert-element-sequence$', user_ajax.insert_element_sequence,
            name='core_composer_insert_element_sequence'),
    re_path(r'^rename-element$', user_ajax.rename_element,
            name='core_composer_rename_element'),
    re_path(r'^delete-element$', user_ajax.delete_element,
            name='core_composer_delete_element'),
    re_path(r'^get-element-occurrences$', user_ajax.get_element_occurrences,
            name='core_composer_get_element_occurrences'),
    re_path(r'^set-element-occurrences$', user_ajax.set_element_occurrences,
            name='core_composer_set_element_occurrences'),
    re_path(r'^save-template$', user_ajax.save_template,
            name='core_composer_save_template'),
    re_path(r'^save-type$', user_ajax.save_type,
            name='core_composer_save_type'),

    re_path(r'^rest/', include('core_composer_app.rest.urls')),
]

urlpatterns = urlpatterns + parser_url



"""AJAX user views of composer application
"""
import json
import logging
from urllib.parse import urlparse

from django.http.response import HttpResponse, HttpResponseBadRequest
from django.template import loader

from core_composer_app.components.type import api as type_api
from core_composer_app.components.type.models import Type
from core_composer_app.components.type_version_manager import api as type_version_manager_api
from core_composer_app.components.type_version_manager.models import TypeVersionManager
from core_composer_app.permissions import rights
from core_composer_app.utils import xml as composer_xml_utils
from core_main_app.commons import exceptions
from core_main_app.components.template.models import Template
from core_main_app.components.template_version_manager import api as template_version_manager_api
from core_main_app.components.template_version_manager.models import TemplateVersionManager
from core_main_app.utils import decorators as decorators
from core_main_app.utils import xml as main_xml_utils
from core_main_app.utils.urls import get_template_download_pattern
from xml_utils.xsd_tree.xsd_tree import XSDTree

logger = logging.getLogger(__name__)


@decorators.permission_required(content_type=rights.composer_content_type,
                                permission=rights.composer_access, raise_exception=True)
def insert_element_sequence(request):
    """Insert the type in the original schema.

    Args:
        request:

    Returns:

    """
    try:
        type_id = request.POST['typeID']
        type_name = request.POST['typeName']
        xpath = request.POST['xpath']
        namespace = request.POST['namespace']
        path = request.POST['path']

        # get link to the type to include
        xsd_string = request.session['newXmlTemplateCompose']

        if type_id == 'built_in_type':
            # insert built-in type into xsd string
            new_xsd_str = composer_xml_utils.insert_element_built_in_type(
                xsd_string, xpath, type_name
            )
        else:
            # get type from database
            type_object = type_api.get(type_id)
            # generate include url
            include_url = main_xml_utils._get_schema_location_uri(str(type_id))
            # insert element in xsd string
            new_xsd_str = composer_xml_utils.insert_element_type(
                xsd_string, xpath, type_object.content, type_name, include_url
            )
            # add the id of the type if not already present
            if include_url not in request.session['includedTypesCompose']:
                request.session['includedTypesCompose'].append(include_url)

        # save the tree in the session
        request.session['newXmlTemplateCompose'] = new_xsd_str

        template = loader.get_template('core_composer_app/user/builder/new_element.html')
        context = {'namespace': namespace,
                   'path': path,
                   'type_name': type_name}
        new_element_html = template.render(context)
        return HttpResponse(
            json.dumps({'new_element': new_element_html}), content_type='application/json'
        )
    except Exception as e:
        return HttpResponseBadRequest(str(e), content_type='application/javascript')


@decorators.permission_required(content_type=rights.composer_content_type,
                                permission=rights.composer_access, raise_exception=True)
def change_xsd_type(request):
    """Change the type of the element.

    Args:
        request:

    Returns:

    """
    try:
        xpath = request.POST['xpath']
        new_type = request.POST['newType']
        xsd_string = request.session['newXmlTemplateCompose']

        # change type
        xsd_string = composer_xml_utils.change_xsd_element_type(xsd_string, xpath, new_type)

        # save the tree in the session
        request.session['newXmlTemplateCompose'] = xsd_string
        return HttpResponse(json.dumps({}), content_type='application/javascript')
    except Exception as e:
        return HttpResponseBadRequest(str(e), content_type='application/javascript')


@decorators.permission_required(content_type=rights.composer_content_type,
                                permission=rights.composer_access, raise_exception=True)
def change_root_type_name(request):
    """Change the name of the root type.

    Args:
        request:

    Returns:

    """
    try:
        type_name = request.POST['typeName']
        xsd_string = request.session['newXmlTemplateCompose']

        # rename root type
        xsd_string = composer_xml_utils.rename_single_root_type(xsd_string, type_name)

        request.session['newXmlTemplateCompose'] = xsd_string
        return HttpResponse(json.dumps({}), content_type='application/javascript')
    except Exception as e:
        return HttpResponseBadRequest(str(e), content_type='application/javascript')


@decorators.permission_required(content_type=rights.composer_content_type,
                                permission=rights.composer_access, raise_exception=True)
def rename_element(request):
    """Replace the current name of the element by the new name.

    Args:
        request:

    Returns:

    """
    try:
        xpath = request.POST['xpath']
        new_name = request.POST['newName']
        xsd_string = request.session['newXmlTemplateCompose']

        # rename element
        xsd_string = composer_xml_utils.rename_xsd_element(xsd_string, xpath, new_name)

        # build xsd tree
        xsd_tree = XSDTree.build_tree(xsd_string)
        # validate the schema
        error = main_xml_utils.validate_xml_schema(xsd_tree)

        if error is not None:
            return _error_response("This is not a valid name.")

        # save the tree in the session
        request.session['newXmlTemplateCompose'] = xsd_string

        return HttpResponse(json.dumps({}), content_type='application/javascript')
    except Exception as e:
        return HttpResponseBadRequest(str(e), content_type='application/javascript')


@decorators.permission_required(content_type=rights.composer_content_type,
                                permission=rights.composer_access, raise_exception=True)
def delete_element(request):
    """Delete an element from the xsd string.

    Args:
        request:

    Returns:

    """
    try:
        xpath = request.POST['xpath']
        xsd_string = request.session['newXmlTemplateCompose']

        # delete element from string
        xsd_string = composer_xml_utils.delete_xsd_element(xsd_string, xpath)

        # save the tree in the session
        request.session['newXmlTemplateCompose'] = xsd_string

        return HttpResponse(json.dumps({}), content_type='application/javascript')
    except Exception as e:
        return HttpResponseBadRequest(str(e), content_type='application/javascript')


@decorators.permission_required(content_type=rights.composer_content_type,
                                permission=rights.composer_access, raise_exception=True)
def get_element_occurrences(request):
    """Get the occurrences of the selected element.

    Args:
        request:

    Returns:

    """
    try:
        xpath = request.POST['xpath']
        xsd_string = request.session['newXmlTemplateCompose']

        # get occurrences of xsd element
        min_occurs, max_occurs = composer_xml_utils.get_xsd_element_occurrences(xsd_string, xpath)

        response_dict = {'minOccurs': min_occurs, 'maxOccurs': max_occurs}
        return HttpResponse(json.dumps(response_dict), content_type='application/javascript')
    except Exception as e:
        return HttpResponseBadRequest(str(e), content_type='application/javascript')


@decorators.permission_required(content_type=rights.composer_content_type,
                                permission=rights.composer_access, raise_exception=True)
def set_element_occurrences(request):
    """Set the occurrences of the selected element.

    Args:
        request:

    Returns:

    """
    try:
        xpath = request.POST['xpath']
        min_occurs = request.POST['minOccurs']
        max_occurs = request.POST['maxOccurs']
        xsd_string = request.session['newXmlTemplateCompose']

        # set element occurrences
        xsd_string = composer_xml_utils.set_xsd_element_occurrences(
            xsd_string, xpath, min_occurs, max_occurs
        )

        # save the tree in the session
        request.session['newXmlTemplateCompose'] = xsd_string
        return HttpResponse(json.dumps({}), content_type='application/javascript')
    except Exception as e:
        return HttpResponseBadRequest(str(e), content_type='application/javascript')


@decorators.permission_required(content_type=rights.composer_content_type,
                                permission=rights.composer_save_template, raise_exception=True)
def save_template(request):
    """Save the current template in the database.

    Args:
        request:

    Returns:

    """
    try:
        template_name = request.POST['templateName']
        xsd_string = request.session['newXmlTemplateCompose']

        response_dict = {}

        try:
            # Build XSD tree
            xsd_tree = XSDTree.build_tree(xsd_string)

            # validate the schema
            error = main_xml_utils.validate_xml_schema(xsd_tree)

            if error is not None:
                return _error_response('This is not a valid XML schema. ' + error)
        except Exception as e:
            return _error_response('This is not a valid XML schema. ' + str(e))

        # get list of dependencies
        dependencies = _get_dependencies_ids(request.session["includedTypesCompose"])

        try:
            # create template version manager
            template_version_manager = TemplateVersionManager(
                title=template_name, user=str(request.user.id)
            )
            # create template
            template = Template(
                filename=template_name, content=xsd_string, dependencies=dependencies
            )
            # save template in database
            template_version_manager_api.insert(template_version_manager, template)
        except exceptions.NotUniqueError:
            return HttpResponseBadRequest(
                "A template with the same name already exists. Please choose another name."
            )
        except Exception as e:
            return _error_response(str(e))

        return HttpResponse(json.dumps(response_dict), content_type='application/javascript')
    except Exception as e:
        return HttpResponseBadRequest(str(e), content_type='application/javascript')


@decorators.permission_required(content_type=rights.composer_content_type,
                                permission=rights.composer_save_type, raise_exception=True)
def save_type(request):
    """Save the current type in the database.

    Args:
        request:

    Returns:

    """
    try:
        type_name = request.POST['typeName']
        template_id = request.POST['templateID']
        xsd_string = request.session['newXmlTemplateCompose']

        response_dict = {}

        # can save as type if new type or from existing type
        if template_id != "new":
            try:
                # check if the type exists, raises exception otherise
                type_api.get(template_id)
            except Exception as e:
                # the type does not exist
                logger.warning("save_type threw an exception: {0}".format(str(e)))
                return _error_response("Unable to save an existing template as a type.")

        try:
            # remove root from tree if present
            xsd_string = composer_xml_utils.remove_single_root_element(xsd_string)
            # build xsd tree
            xsd_tree = XSDTree.build_tree(xsd_string)
            # validate the schema
            error = main_xml_utils.validate_xml_schema(xsd_tree)

            if error is not None:
                return _error_response('This is not a valid XML schema. ' + error)
        except Exception as e:
            return _error_response('This is not a valid XML schema. ' + str(e))

        dependencies = _get_dependencies_ids(request.session["includedTypesCompose"])

        try:
            # create type version manager
            type_version_manager = TypeVersionManager(title=type_name, user=str(request.user.id))
            # create type
            type_object = Type(filename=type_name, content=xsd_string, dependencies=dependencies)
            # save type in database
            type_version_manager_api.insert(type_version_manager, type_object)
        except exceptions.NotUniqueError:
            return HttpResponseBadRequest(
                "A type with the same name already exists. Please choose another name."
            )
        except Exception as e:
            return _error_response(str(e))

        return HttpResponse(json.dumps(response_dict), content_type='application/javascript')
    except Exception as e:
        return HttpResponseBadRequest(str(e), content_type='application/javascript')


def _get_dependencies_ids(list_dependencies):
    """Return list of type ids from list of dependencies.

    Args:
        list_dependencies:

    Returns:

    """
    # declare list of dependencies
    dependencies = []
    # get pattern to match a template download
    pattern = get_template_download_pattern()
    # get all type ids
    for uri in list_dependencies:
        # parse dependency url
        url = urlparse(uri)
        try:
            # get object id from url
            object_id = pattern.match(url.path).group('pk')
            # get type by id, exception raised if not found
            type_object = type_api.get(object_id)
            # add id to list of internal dependencies
            dependencies.append(type_object)
        except Exception as e:
            # id not found, don't add it to list of dependencies
            logger.warning("_get_dependencies_ids threw an exception: {0}".format(str(e)))

    return dependencies


def _error_response(error):
    """Return HttpResponse containing the error message.

    Args:
        error:

    Returns:

    """
    return HttpResponseBadRequest(error.replace("'", ""), content_type='application/javascript')



"""Url router for the administration site
"""
from django.conf import settings
from django.contrib import admin
from django.urls import re_path, reverse_lazy

from core_composer_app.views.admin import views as admin_views, ajax as admin_ajax
from core_composer_app.views.admin.ajax import EditBucketView
from core_main_app.views.common.ajax import EditTemplateVersionManagerView

parser_url = []
if 'core_parser_app' in settings.INSTALLED_APPS:
    from core_parser_app.views.admin import views as admin_parser_views

    parser_url = [
        re_path(r'^type/modules/(?P<pk>\w+)',
                admin_parser_views.ManageModulesAdminView.as_view(
                    back_to_previous_url="admin:core_composer_app_manage_type_versions"
                ),
                name='core_composer_app_type_modules'),
    ]

admin_urls = [
    re_path(r'^types$', admin_views.manage_types,
            name='core_composer_app_types'),
    re_path(r'^type/(?P<pk>[\w-]+)/edit/$',
            EditTemplateVersionManagerView.as_view(success_url=reverse_lazy(
                "admin:core_composer_app_types")),
            name='core_composer_app_edit_type'),
    re_path(r'^type/upload$', admin_views.upload_type,
            name='core_composer_app_upload_type'),
    re_path(r'^type/upload/(?P<version_manager_id>\w+)', admin_views.upload_type_version,
            name='core_composer_app_upload_type_version'),
    re_path(r'^type/versions/(?P<version_manager_id>\w+)', admin_views.manage_type_versions,
            name='core_composer_app_manage_type_versions'),
    re_path(r'^type/buckets/(?P<version_manager_id>\w+)', admin_views.manage_type_buckets,
            name='core_composer_app_type_buckets'),
    re_path(r'^buckets$', admin_views.manage_buckets,
            name='core_composer_app_buckets'),
    re_path(r'^bucket/upload$', admin_views.upload_bucket,
            name='core_composer_app_upload_bucket'),
    re_path(r'^bucket/delete$', admin_ajax.delete_bucket,
            name='core_composer_app_delete_bucket'),
    re_path(r'^type/resolve-dependencies', admin_ajax.resolve_dependencies,
            name='core_composer_app_resolve_dependencies'),
    re_path(r'^bucket/(?P<pk>[\w-]+)/edit/$',
            EditBucketView.as_view(),
            name='core_composer_app_edit_bucket'),
]

urls = admin.site.get_urls()
admin.site.get_urls = lambda: admin_urls + urls + parser_url



"""
Type API
"""
from core_composer_app.components.type.models import Type
from core_composer_app.utils.xml import check_type_core_support, COMPLEX_TYPE
from core_main_app.components.template import api as template_api


def upsert(type_object):
    """Save or update the type.

    Args:
        type_object:

    Returns:

    """
    # Check that the type is supported by the core
    type_definition = check_type_core_support(type_object.content)
    type_object.is_complex = type_definition == COMPLEX_TYPE
    # Save type
    return template_api.upsert(type_object)


def get(type_id):
    """Get a type.

    Args:
        type_id:

    Returns:

    """
    return Type.get_by_id(type_id)


def get_all():
    """List all types.

    Returns:

    """
    return Type.get_all()


def get_all_complex_type():
    """List all complex types.

    Returns:

    """
    return Type.get_all_complex_type()


def delete(type_object):
    """Delete the type.

    Returns:

    """
    type_object.delete()



"""Bucket model
"""
from django_mongoengine import fields, Document
from mongoengine import errors as mongoengine_errors

from core_composer_app.components.type_version_manager.models import TypeVersionManager
from core_main_app.commons import exceptions


class Bucket(Document):
    """Bucket class to store types by domain.
    """
    label = fields.StringField(unique=True)
    color = fields.StringField(unique=True)
    types = fields.ListField(fields.ReferenceField(TypeVersionManager), blank=True)

    @staticmethod
    def get_by_id(bucket_id):
        """Return a bucket given its id.

        Args:
            bucket_id:

        Returns:

        """
        try:
            return Bucket.objects.get(pk=str(bucket_id))
        except mongoengine_errors.DoesNotExist as e:
            raise exceptions.DoesNotExist(str(e))
        except Exception as ex:
            raise exceptions.ModelError(str(ex))

    @staticmethod
    def get_all():
        """Return all buckets.

        Returns:

        """
        return Bucket.objects().all()

    @staticmethod
    def get_colors():
        """Return all colors.

        Returns:

        """
        return Bucket.objects.values_list('color')

    def save_object(self):
        """Custom save

        Returns:

        """
        try:
            return self.save()
        except mongoengine_errors.NotUniqueError as e:
            raise exceptions.NotUniqueError(str(e))
        except Exception as ex:
            raise exceptions.ModelError(str(ex))



""" REST abstract views for the type version manager API
"""
from abc import ABCMeta, abstractmethod

from django.http import Http404
from rest_framework import status
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response
from rest_framework.views import APIView

from core_composer_app.rest.type_version_manager.serializers import CreateTypeSerializer, TypeVersionManagerSerializer
from core_main_app.commons.exceptions import NotUniqueError, XSDError
from core_main_app.rest.template_version_manager.abstract_views import AbstractTemplateVersionManagerDetail


class AbstractTypeList(APIView, metaclass=ABCMeta,):
    """ Create a type & type version manager
    """

    def post(self, request):
        """ Create a type & type version manager

        Parameters:

            {
                "title": "title",
                "filename": "filename",
                "content": "<xs:schema xmlns:xs='http://www.w3.org/2001/XMLSchema'><xs:simpleType name='root'/></xs:schema>"
            }

        Note:

            "dependencies_dict": json.dumps({"schemaLocation1": "id1" ,"schemaLocation2":"id2"})

        Args:

            request: HTTP request

        Returns:

            - code: 201
              content: Type
            - code: 400
              content: Validation error / bad request
            - code: 500
              content: Internal server error
        """
        try:
            # Build serializers
            type_serializer = CreateTypeSerializer(data=request.data)
            type_version_manager_serializer = TypeVersionManagerSerializer(data=request.data)

            # Validate data
            type_serializer.is_valid(True)
            type_version_manager_serializer.is_valid(True)

            # Save data
            type_version_manager_object = type_version_manager_serializer.save(user=self.get_user())
            type_serializer.save(type_version_manager=type_version_manager_object)

            return Response(type_serializer.data, status=status.HTTP_201_CREATED)
        except ValidationError as validation_exception:
            content = {'message': validation_exception.detail}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except NotUniqueError:
            content = {'message': "A type with the same title already exists."}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except XSDError as xsd_error:
            content = {'message': "XSD Error: " + str(xsd_error)}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except Exception as api_exception:
            content = {'message': str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    @abstractmethod
    def get_user(self):
        """ Get a user
        """
        raise NotImplementedError("get_user method is not implemented.")


class TypeVersion(AbstractTemplateVersionManagerDetail):
    """ Create a type version
    """

    def post(self, request, pk):
        """ Create a type version

        Parameters:

            {
                "filename": "filename",
                "content": "<xs:schema xmlns:xs='http://www.w3.org/2001/XMLSchema'><xs:simpleType name='root'/></xs:schema>"
            }

        Note:

            "dependencies_dict": json.dumps({"schemaLocation1": "id1" ,"schemaLocation2":"id2"})

        Args:

            request: HTTP request

        Returns:

            - code: 201
              content: Type
            - code: 400
              content: Validation error / bad request
            - code: 404
              content: Object was not found
            - code: 500
              content: Internal server error
        """
        try:
            # Get object
            type_version_manager_object = self.get_object(pk)

            # Build serializers
            type_serializer = CreateTypeSerializer(data=request.data)

            # Validate data
            type_serializer.is_valid(True)

            # Save data
            type_serializer.save(type_version_manager=type_version_manager_object)

            return Response(type_serializer.data, status=status.HTTP_201_CREATED)
        except Http404:
            content = {'message': 'Type Version Manager not found.'}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except ValidationError as validation_exception:
            content = {'message': validation_exception.detail}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except Exception as api_exception:
            content = {'message': str(api_exception)}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)
